import logging
from collections import namedtuple

from .exception_util import exception_as_string
from .conditional_sleep import sleep_with_condition
from .value_to_bool import value_to_bool

__all__ = (
    "exception_as_string",
    "async_util",
    "caching",
    "sleep_with_condition",
    "hibp",
    "value_to_bool",
    "timing",
)
__version__ = "1.5.1"
VersionInfo = namedtuple("VersionInfo", "major minor micro releaselevel serial")
version_info = VersionInfo(major=1, minor=5, micro=1, releaselevel="final", serial=0)
logging.getLogger(__name__).addHandler(logging.NullHandler())



import asyncio
import hashlib
import logging
import sys

import commons

try:
    import httpx  # noqa
except ImportError:
    # So we can continue to ship zero deps
    print("httpx is required to use this, please install it and try again.")
    sys.exit(1)

log = logging.getLogger(__name__)


async def has_password_been_pwned(password: str) -> bool:
    """Lookup the password in 'have i been pwned'

    Returns
    -------
    bool
        True if pwned, false in all other cases

    Notes
    -----
    Fails safely in the sense that a failed
    lookup returns False
    """
    password_hash = hashlib.sha1(password.encode()).hexdigest().upper()
    first_chars = password_hash[:5]
    remaining_chars = password_hash[5:]
    assert first_chars + remaining_chars == password_hash
    try:
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"https://api.pwnedpasswords.com/range/{first_chars}",
                timeout=5,
            )
    except Exception as e:
        log.critical(
            "Pwned passwords lookup failed:\n%s", commons.exception_as_string(e)
        )
        return False

    if remaining_chars in response.text:
        return True

    return False


async def main():
    print(await has_password_been_pwned("test"))
    print(await has_password_been_pwned("qwerty"))
    print(await has_password_been_pwned("fbghbdajkwdjkwadkwadjkwad"))


if __name__ == "__main__":
    asyncio.run(main())



import asyncio


async def maybe_coro(func, *args, **kwargs):
    """Call the given func, awaiting if required.

    Args and Kwargs are passed directly to the coroutine.
    """
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)

    return func(*args, **kwargs)



import asyncio
import logging
import random
from typing import Any

log = logging.getLogger(__name__)
background_tasks: set[tuple[Any, asyncio.Task]] = set()


def create_task(*args, task_id=None, **kwargs) -> asyncio.Task:
    """Creates a task while maintain a background reference, so they don't get de-reffed

    Parameters
    ----------
    task_id
        A unique variable to store tasks under

        If None, uses random.randbytes(4)
    """
    if task_id is None:
        task_id = random.randbytes(4)

    task = asyncio.create_task(*args, **kwargs)
    background_tasks.add((task_id, task))
    task.add_done_callback(lambda _: background_tasks.discard((task_id, task)))
    return task



import asyncio
import functools
from typing import Union, Callable, Any, Coroutine


async def sleep_with_condition(
    seconds: float,
    condition: Union[
        functools.partial,
        Callable[[Any], bool],
        Callable[[Any], Coroutine[Any, Any, bool]],
    ],
    *,
    interval: float = 5,
) -> None:
    """Sleep until either condition is True or we run out of seconds

    Parameters
    ----------
    seconds: float
        How long to sleep for up to
    condition
        Pass either:
        - A sync function to call which returns a bool
        - An async function to call which returns a bool

        .. note::

            If you wish to use arguments in you functions,
            pass an instance of :class:`functools.partial`
    interval: float
        How long to sleep in-between each condition check.

        Defaults to 5 seconds.
    """
    if asyncio.iscoroutinefunction(condition):
        wrapped_condition = condition
    elif callable(condition) or isinstance(condition, functools.partial):

        async def wrapped_condition():
            return condition()

    else:
        raise TypeError("Unknown input argument")

    remaining_seconds = seconds
    while remaining_seconds > 0:
        remaining_seconds -= interval
        await asyncio.sleep(interval)

        if await wrapped_condition():
            return



import datetime


def get_current_utc() -> datetime.datetime:
    """Fetch a UTC timezone aware datetime.

    Returns
    -------
    datetime.datetime
        The current UTC time as an aware datetime.
    """
    return datetime.datetime.now(tz=datetime.timezone.utc)


def is_within_next_(
    current_datetime: datetime.datetime,
    expected_datetime: datetime.datetime,
    delta: datetime.timedelta = datetime.timedelta(minutes=5),
) -> bool:
    """Returns true if the times are within X minutes of each other

    Parameters
    ----------
    current_datetime : datetime.datetime
        Now.
    expected_datetime : datetime.datetime
        When you want the thing to occur.
    delta : datetime.timedelta
        Return true if the two datetimes are within this delta.

        Defaults to 5 minutes.

    Returns
    -------
    bool
        True if times within the given delta.
    """
    if is_in_the_past(current_datetime, expected_datetime):
        # Date is in the past
        return False

    difference = expected_datetime - current_datetime
    if difference < delta:
        return True

    return False


def is_in_the_past(
    current_datetime: datetime.datetime,
    expected_datetime: datetime.datetime,
) -> bool:
    """Returns True if the datetime is in the past."""
    return current_datetime > expected_datetime



def value_to_bool(value: str | bool | None) -> bool:
    """Convert a string representation of truth to true (1) or false (0).
    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.

    Also handles the case the value is a bool or None

    Raises ValueError if 'val' is anything else.
    """
    if value is None or value == "":
        return False

    if isinstance(value, bool):
        return value

    if isinstance(value, int):
        value = str(value)

    value = value.lower()
    if value in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif value in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError("invalid truth value %r" % (value,))



from dataclasses import dataclass
from datetime import timedelta, datetime
from typing import Any, Dict, Optional, Generic, TypeVar

KT = TypeVar("KT", bound=Any)
VT = TypeVar("VT", bound=Any)


class ExistingEntry(Exception):
    """An entry was already found in the cache with this key."""


class NonExistentEntry(Exception):
    """No entry found in the cache with this key."""


@dataclass(slots=True)
class Entry:
    value: Any
    expiry_time: Optional[datetime] = None


class TimedCache(Generic[KT, VT]):
    __slots__ = ("cache", "global_ttl", "non_lazy", "ttl_from_last_access")

    def __init__(
        self,
        *,
        global_ttl: Optional[timedelta] = None,
        lazy_eviction: bool = True,
        ttl_from_last_access: bool = False,
    ):
        """
        Parameters
        ----------
        global_ttl: Optional[timedelta]
            A default TTL for any added entries.
        lazy_eviction: bool
            Whether this cache should perform lazy eviction or not.

            Defaults to True
        ttl_from_last_access: bool
            Whether the TTL of an object is dictated by
            the last access or time of insertion.

            This requires a global TTL to be set.

            Defaults to True, time of insertion.
        """
        self.cache: Dict[KT, Entry] = {}
        self.non_lazy: bool = not lazy_eviction
        self.global_ttl: Optional[timedelta] = global_ttl
        self.ttl_from_last_access: bool = ttl_from_last_access

        if self.ttl_from_last_access is True and self.global_ttl is None:
            raise ValueError(
                "Cannot set ttl_from_last_access without global_ttl also being set."
            )

    def __contains__(self, item: Any) -> bool:
        try:
            entry = self.cache[item]
            if entry.expiry_time and entry.expiry_time < datetime.now():
                self.delete_entry(item)
                return False
        except KeyError:
            return False
        else:
            return True

    def __len__(self):
        self.force_clean()
        return len(self.cache.keys())

    def add_entry(
        self,
        key: KT,
        value: VT,
        *,
        ttl: Optional[timedelta] = None,
        override: bool = False,
    ) -> None:
        """
        Add an entry to the cache.

        Parameters
        ----------
        key
            The key to store this under.
        value
            The item you want to store in the cache
        ttl: Optional[timedelta]
            An optional period of time to expire
            this entry after.
        override: bool
            Whether or not to override an existing value

        Raises
        ------
        ExistingEntry
            You are trying to insert a duplicate key

        Notes
        -----
        ttl passed to this method will
        take precendence over the global ttl.
        """
        self._perform_eviction()
        if key in self and not override:
            raise ExistingEntry

        if ttl or self.global_ttl:
            ttl = ttl or self.global_ttl
            self.cache[key] = Entry(value=value, expiry_time=(datetime.now() + ttl))
        else:
            self.cache[key] = Entry(value=value)

    def delete_entry(self, key: KT) -> None:
        """
        Delete a key from the cache

        Parameters
        ----------
        key
            The key to delete
        """
        self._perform_eviction()
        try:
            self.cache.pop(key)
        except KeyError:
            pass

    def get_entry(self, key: KT) -> VT:
        """
        Fetch a value from the cache

        Parameters
        ----------
        key
            The key you wish to
            retrieve a value for

        Returns
        -------
        VT
            The provided value

        Raises
        ------
        NonExistentEntry
            No value exists in the cache
            for the provided key.

        """
        self._perform_eviction()
        if key not in self:
            raise NonExistentEntry

        entry: Entry = self.cache[key]
        if self.ttl_from_last_access and entry.expiry_time is not None:
            if self.global_ttl is None:
                raise ValueError(
                    "ttl_from_last_access requires global_ttl also being set."
                )

            entry.expiry_time = datetime.now() + self.global_ttl

        return entry.value

    def force_clean(self) -> None:
        """
        Clear out all outdated cache items.
        """
        now = datetime.now()
        self.cache = {
            k: v
            for k, v in self.cache.items()
            if (v.expiry_time and v.expiry_time > now) or not v.expiry_time
        }

    def _perform_eviction(self):
        if self.non_lazy:
            self.force_clean()
